/**
  *  \file planetscentral/host/rating.cc
  *  \brief Host Server: Difficulty Ratings
  *
  *  These functions rate a game or tool. Unlike the standalone rating tool, c2rater,
  *  this has to deal with two obstacles:
  *  - tools have potentially totally incomplete configuration. We still must produce
  *    a guess how it will affect the rating. Therefore, DifficultyRater is able to
  *    produce individual ratings when given enough configuration to compute them.
  *    We therefore feed the DifficultyRater with partial configuration.
  *  - for a game, tools can override the game difficulty. For unmastered games, this
  *    is pretty easy: if the tool specifies a difficulty, we exclude it from the
  *    rating computation. For mastered games, this is hard because we don't know
  *    where the actual config values come from. A tool that configures a NativeRating
  *    of 150 and specifies an override of 180 would be rated as 150%*180% = 270%.
  *    Overrides therefore only work reliably for two cases:
  *    - ship lists. Here we know how to exclude values: if a ship list specifies a
  *      rating, we don't rate the ship list files.
  *    - tools that affect configuration which is not rated by the normal formula.
  */

#include <iostream>
#include <stdio.h>
#include "planetscentral/host/rating.h"
#include "cpluslib/misc.h"
#include "game/formula.h"
#include "io/dirs.h"
#include "io/streams.h"
#include "planetscentral/host/db.h"
#include "planetscentral/host/game.h"
#include "planetscentral/host/host.h"
#include "planetscentral/osutil.h"
#include "planetscentral/rater/difficulty.h"
#include "util/except.h"

namespace {
    enum {
        HaveAMaster = 1,
        HavePMaster = 2
    };
}

static bool
processDirectoryFile(DifficultyRater& rater, Directory& dir, const char* name)
{
    if (Ptr<Stream> s = dir.openFileNT(name, Stream::C_READ)) {
        rater.addConfigFile(*s);
        return true;
    } else {
        return false;
    }
}

static void
processDirectory(DifficultyRater& rater, uint32_t& state, string_t dir)
{
    // No directory (could be tool without directory): ignore
    if (dir.empty()) {
        return;
    }

    // Read ship list unless we already have one
    Ptr<Directory> d = Directory::openDir(dir);
    if (!rater.isRatingKnown(DifficultyRater::ShiplistRating)) {
        rater.addShipList(*d);
    }

    // Read configuration files
    processDirectoryFile(rater, *d, "pconfig.src");
    processDirectoryFile(rater, *d, "pconfig.src.frag");
    processDirectoryFile(rater, *d, "shiplist.txt");
    processDirectoryFile(rater, *d, "shiplist.txt.frag");

    // Conditionally process master config files. Do not process
    // AMaster config if we have a PMaster config and vice versa;
    // only process matching fragments.
    if ((state & (HaveAMaster|HavePMaster)) == 0 && processDirectoryFile(rater, *d, "amaster.src")) {
        state |= HaveAMaster;
    }
    if ((state & (HaveAMaster|HavePMaster)) == 0 && processDirectoryFile(rater, *d, "pmaster.cfg")) {
        state |= HavePMaster;
    }
    if ((state & HaveAMaster) != 0) {
        processDirectoryFile(rater, *d, "amaster.src.frag");
    }
    if ((state & HavePMaster) != 0) {
        processDirectoryFile(rater, *d, "pmaster.cfg.frag");
    }
}

static bool
processRatingForTool(double& modifier, DbSubtree tree, string_t toolName)
{
    DbHash tool(tree.subtree("prog").hashKey(toolName));
    bool result = false;
    if (tool.getIntField("useDifficulty")) {
        modifier *= tool.getIntField("difficulty");
        modifier /= 100.0;
        result = true;
    }
    return result;
}

static void
processDirectoryForTool(double& modifier, DifficultyRater& rater, uint32_t& state, DbSubtree tree, string_t toolName)
{
    if (!processRatingForTool(modifier, tree, toolName)) {
        string_t toolDir = tree.subtree("prog").hashKey(toolName).getStringField("path");
        processDirectory(rater, state, toolDir);
    }
}

/** Format computed rating to external format.
    \param rating Rating, floating point, 1.0 for standard rating
    \return [1,1000], 100 for standard rating */
static int
formatRating(double rating)
{
    int result = roundToInt(100 * rating);
    if (result < 1) {
        result = 1;
    }
    if (result > 1000) {
        result = 1000;
    }
    return result;
}

/** Compute rating of a tool, given a directory.
    \param dir Directory
    \return rating, 1..100 */
int
computeToolRating(string_t dir)
{
    try {
        Ptr<Directory> d = Directory::openDir(dir);

        // Assume it is a ship list and rate as such
        DifficultyRater rater;
        rater.addShipList(*d);

        // Configuration files. Note that this is a close relative to addConfigDirectory().
        // It is important that we process the ".frag" after the original files.
        // It is not a problem if we process both AMaster and PMaster config files, as
        // they're expected to have the same content. A problem would be a tool containing
        // both amaster.src and pmaster.cfg.frag (but no amaster.src.frag and pmaster.cfg),
        // but this doesn't happen with any sane tool.
        static const char*const files[] = {
            "amaster.src",  "amaster.src.frag",
            "pconfig.src",  "pconfig.src.frag",
            "pmaster.cfg"   "pmaster.cfg.frag"
            "shiplist.txt", "shiplist.txt.frag",
        };
        for (uint_t i = 0; i < countof(files); ++i) {
            Ptr<Stream> s = d->openFileNT(files[i], Stream::C_READ);
            if (s) {
                rater.addConfigFile(*s);
            }
        }
        return formatRating(rater.getTotalRating());
    }
    catch (FileProblemException& e) {
        MutexGuard g(output_mutex);
        printf("[rating] tool rating '%s': %s\n", dir.c_str(), e.what());
        return 100;
    }
}

/** Compute rating of a game.
    \param conn Database connection
    \param g Game */
int
computeGameRating(DbClient& conn, Game& g)
{
    try {
        string_t gameDir(g.getDirectory() + "/data");
        DifficultyRater rater;
        StringList_t tools;
        g.getGame().hashKey("toolkind").getAll(tools);

        double modifier = 1.0;
        if (g.getConfigInt("masterHasRun")) {
            // Master has run, so the game directory is fully-populated.
            // We still have to check whether any component enforces a particular
            // rating modifier.
            Ptr<Directory> d = Directory::openDir(gameDir);
            processRatingForTool(modifier, DbSubtree(conn, HOST_PROGRAM_ROOT),     g.getConfig("host"));
            processRatingForTool(modifier, DbSubtree(conn, MASTER_PROGRAM_ROOT),   g.getConfig("master"));
            bool slmod = processRatingForTool(modifier, DbSubtree(conn, SHIPLIST_PROGRAM_ROOT), g.getConfig("shiplist"));
            for (uint32_t i = 0; i+1 < tools.size(); i += 2) {
                processRatingForTool(modifier, DbSubtree(conn, TOOL_PROGRAM_ROOT), tools[i+1]);
            }
            if (!slmod) {
                rater.addShipList(*d);
            }
            rater.addConfigDirectory(*d);
        } else {
            // Master has not run, so gather data from individual directories
            uint32_t state = 0;
            processDirectory(rater, state, gameDir);
            processDirectoryForTool(modifier, rater, state, DbSubtree(conn, HOST_PROGRAM_ROOT),     g.getConfig("host"));
            processDirectoryForTool(modifier, rater, state, DbSubtree(conn, MASTER_PROGRAM_ROOT),   g.getConfig("master"));
            processDirectoryForTool(modifier, rater, state, DbSubtree(conn, SHIPLIST_PROGRAM_ROOT), g.getConfig("shiplist"));
            for (uint32_t i = 0; i+1 < tools.size(); i += 2) {
                processDirectoryForTool(modifier, rater, state, DbSubtree(conn, TOOL_PROGRAM_ROOT), tools[i+1]);
            }
        }
        return formatRating(rater.getTotalRating() * modifier);
    }
    catch (FileProblemException& e) {
        MutexGuard gg(output_mutex);
        printf("[rating] game rating #%d: %s\n", int(g.getId()), e.what());
        return 100;
    }
}
