/**
  *  \file planetscentral/talk/bbparse.cc
  *  \brief Talk: BBCode parser
  */

#include <cstring>
#include "planetscentral/talk/bbparse.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "util/misc.h"
#include "cpluslib/format.h"

static const BBColor colors[] = {
    // Some usual colors
    { "black",                     "#000000" },
    { "red",                       "#ff0000" },
    { "green",                     "#00ff00" },
    { "blue",                      "#0000ff" },
    { "white",                     "#ffffff" },
    { "yellow",                    "#ffff00" },
    { "magenta",                   "#ff00ff" },
    { "cyan",                      "#00ffff" },

    // PCC colors
    { "player1",                   "#808096" },
    { "player2",                   "#ffffff" },
    { "player3",                   "#ffff00" },
    { "player4",                   "#61f261" },
    { "player5",                   "#6161c2" },
    { "player6",                   "#ff0000" },
    { "player7",                   "#ff55ff" },
    { "player8",                   "#c26100" },
    { "player9",                   "#ffc200" },
    { "player10",                  "#55ffff" },
    { "player11",                  "#00aa00" },
};

const BBColor*
BBColor::findColorByName(const char* name)
{
    for (std::size_t i = 0; i < countof(colors); ++i) {
        if (std::strcmp(colors[i].name, name) == 0) {
            return &colors[i];
        }
    }
    return 0;
}

const BBColor*
BBColor::findColorByValue(const char* rgb)
{
    for (std::size_t i = 0; i < countof(colors); ++i) {
        if (std::strcmp(colors[i].rgb, rgb) == 0) {
            return &colors[i];
        }
    }
    return 0;
}

namespace {
    struct Tag {
        const char* name;
        uint8_t major;
        uint8_t minor;
    };
}

/** Tags. All tags that are handled in a pretty regular fashion.
    Notable exclusions:
    - [noparse] which has no tree equivalent and is handled ad-hoc
    - [code] which is handled ad-hoc
    - [*] which has no close tag */
static const Tag tags[] = {
    { "b",      maInline,     miInBold },
    { "i",      maInline,     miInItalic },
    { "s",      maInline,     miInStrikeThrough },
    { "strike", maInline,     miInStrikeThrough },
    { "u",      maInline,     miInUnderline },
    { "tt",     maInline,     miInMonospace },
    { "color",  maInlineAttr, miIAColor },
    { "size",   maInlineAttr, miIASize },
    { "font",   maInlineAttr, miIAFont },
    { "url",    maLink,       miLinkUrl },
    { "email",  maLink,       miLinkEmail },
    { "thread", maLink,       miLinkThread },
    { "post",   maLink,       miLinkPost },
    { "game",   maLink,       miLinkGame },
    { "forum",  maLink,       miLinkForum },
    { "user",   maLink,       miLinkUser },
    { "img",    maSpecial,    miSpecialImage },
    { "center", maParagraph,  miParCentered },
    { "quote",  maGroup,      miGroupQuote },
    { "list",   maGroup,      miGroupList },
};

static const Tag*
identify(const string_t& tag)
{
    for (uint_t i = 0; i != countof(tags); ++i) {
        if (tag == tags[i].name) {
            return &tags[i];
        }
    }
    return 0;
}

/** Convert a maInlineAttr attribute to canonical format. */
static bool
canonicalizeAttribute(uint8_t kind, const string_t& in, string_t& out)
{
    // An empty attribute is never valid
    string_t t(strTrim(in));
    if (t.empty()) {
        return false;
    }

    // Type-dependant processing
    int delta;
    static const char digits[] = "0123456789abcdef";
    switch (kind) {
     case miIAColor:
        // We accept a number of formats:
        // 1. "#rrggbb"
        // 2. "#rgb"
        // 3. "rrggbb"
        // 4. "rgb"
        // 5. named
        // Output always is "#rrggbb".
        t = strLCase(t);
        if (t[0] == '#' && t.size() == 7 && t.find_first_not_of(digits, 1) == t.npos) {
            out = t;
            return true;
        } else if (t[0] == '#' && t.size() == 4 && t.find_first_not_of(digits, 1) == t.npos) {
            out = "#";
            out += t[1]; out += t[1];
            out += t[2]; out += t[2];
            out += t[3]; out += t[3];
            return true;
        } else if (t.size() == 6 && t.find_first_not_of(digits) == t.npos) {
            out = "#";
            out += t;
            return true;
        } else if (t.size() == 3 && t.find_first_not_of(digits) == t.npos) {
            out = "#";
            out += t[0]; out += t[0];
            out += t[1]; out += t[1];
            out += t[2]; out += t[2];
            return true;
        } else if (const BBColor* c = BBColor::findColorByName(t.c_str())) {
            out = c->rgb;
            return true;
        } else {
            return false;
        }

     case miIASize:
        // We accept numeric values. Constants are translated into relative sizes, assuming that
        // the default size is 5. Whatever unit that might be.
        if (stringValue(in, delta)) {
            if (in[0] != '-' && in[0] != '+') {
                delta -= 5;
            }
            if (delta >= -8 && delta <= +8) {
                out = format("%+d", delta);
                return true;
            } else {
                return false;
            }
        } else {
            return false;
        }

     case miIAFont:
        // Fonts are taken verbatim, but we refuse characters that we'll most likely
        // not be able to quote
        if (t.find("\"';\\\n/") != in.npos) {
            return false;
        }
        out = t;
        return true;

     default:
        return false;
    }
}

/** Complete a link node.
    If the user didn't specify the link as the attribute, derive it from the content. */
static void
completeLink(TextNode* node)
{
    if (node->text.empty()) {
        // There is no target attribute. Get it from the content.
        if (node->children.size() == 1 && node->children[0]->major == maPlain) {
            // It contains just plain text, as in "[user]fruno[/user]".
            // This will give it an empty content, so the renderer will
            // generate it anew.
            node->text = strTrim(node->children[0]->text);
            node->children.pop_back();
        } else {
            // The content has formatting, as in "[user][b]f[/b]runo[/user]".
            // Assume the user wants to keep that, and just use the raw text.
            node->text = strTrim(node->children[0]->rawTextContent());
        }
    }
}


BBParser::BBParser(BBLexer& lex, InlineItemSet options)
    : lex(lex),
      options(options),
      current(),
      stack()
{
    open(maGroup, miGroupRoot);
    open(maParagraph, miParNormal);
}

/*
 *  Invariants: The stack always consists of
 *  - 0 or more groups (maGroup)
 *  - one paragraph (maParagraph)
 *  - 0 or more inline markups (maInline, maInlineAttr, maLink)
 *
 *  There never is a plain-text node on the stack.
 */

TextNode*
BBParser::parse()
{
    next();
    while (current != BBLexer::Eof) {
        switch (current) {
         case BBLexer::TagStart:  handleStart();     break;
         case BBLexer::TagEnd:    handleEnd();       break;
         case BBLexer::Smiley:    handleSmiley();    break;
         case BBLexer::Paragraph: handleParagraph(); break;
         case BBLexer::AtLink:    handleAtLink();    break;
         case BBLexer::Text:      handleText();      break;
         case BBLexer::Eof:                          break;
        }
        next();
    }
    close(1);
    return stack.extract_last();
}

/** Check for known tag. */
bool
BBParser::isKnownTag(const string_t& tag)
{
    // xref handleStart()
    return tag == "noparse"
        || tag == "code"
        || tag == "*"
        || tag == "break"
        || tag == "nl"
        || identify(tag) != 0;
}

/** Handle a start tag. */
void
BBParser::handleStart()
{
    // xref isKnownTag()
    if (lex.getTag() == "noparse") {
        handleNoparse();
    } else if (lex.getTag() == "code") {
        handleCode();
    } else if (lex.getTag() == "*") {
        handleListItem();
    } else if (lex.getTag() == "break") {
        // Close inline and paragraph, insert a break, open a paragraph
        closeInline();
        close();
        open(maParagraph, miParBreak);
        close();
        open(maParagraph, miParNormal);
        lex.skipBlanks();
    } else if (lex.getTag() == "nl") {
        // Line break
        open(maSpecial, miSpecialBreak);
        close();
    } else if (const Tag* t = identify(lex.getTag())) {
        switch (TextMajorKind(t->major)) {
         case maInline:
            // Just open it.
            open(t->major, t->minor);
            break;

         case maInlineAttr:
            // Just open it if attribute allows.
            {
                string_t attr;
                if (canonicalizeAttribute(t->minor, lex.getAttribute(), attr)) {
                    open(t->major, t->minor);
                    stack.back()->text = attr;
                } else {
                    appendText(lex.getTokenString());
                }
            }
            break;

         case maLink:
            // Open it, but make sure we don't nest it.
            closeLinks();
            open(t->major, t->minor);
            stack.back()->text = lex.getAttribute();
            break;

         case maParagraph:
            // Close inline, open paragraph
            closeInline();
            close();
            open(t->major, t->minor);
            lex.skipBlanks();
            break;

         case maGroup:
            if (t->minor == miGroupList) {
                // Close inline, open the list
                closeInline();
                close();
                open(maGroup, miGroupList);
                stack.back()->text = lex.getAttribute();
                open(maGroup, miGroupListItem);
                open(maParagraph, miParNormal);
                lex.skipBlanks();
            } else if (t->minor == miGroupQuote) {
                // Close inline, open the quote
                closeInline();
                close();
                open(maGroup, miGroupQuote);
                stack.back()->text = lex.getAttribute();
                open(maParagraph, miParNormal);
                lex.skipBlanks();
            } else {
                // Can't happen
                appendText(lex.getTokenString());
            }
            break;

         case maSpecial:
            // Usually an image. Just open it, but don't nest.
            closeUntil(t->major, t->minor);
            open(t->major, t->minor);
            stack.back()->text = lex.getAttribute();
            break;

         case maPlain:
            // Can't happen
            appendText(lex.getTokenString());
            break;
        }
    } else {
        // Unknown
        appendText(lex.getTokenString());
    }
}

/** Handle an end tag. */
void
BBParser::handleEnd()
{
    const Tag* t = identify(lex.getTag());
    if (t == 0) {
        appendText(lex.getTokenString());
    } else {
        switch (TextMajorKind(t->major)) {
         case maInline:
         case maInlineAttr:
         case maLink:
         case maSpecial:
            // Inline formatting.
            {
                depth_t i = stack.size()-1;
                while (i > 0
                       && (stack[i]->major != t->major || stack[i]->minor != t->minor)
                       && (stack[i]->major == maInline
                           || stack[i]->major == maInlineAttr
                           || stack[i]->major == maLink
                           || (stack[i]->major == maSpecial && stack[i]->minor == miSpecialImage)))
                {
                    --i;
                }
                if (i == 0) {
                    // Cannot close it (can't happen)
                    appendText(lex.getTokenString());
                } else if (stack[i]->major == t->major && stack[i]->minor == t->minor) {
                    // We can close it
                    close(i);
                } else if (stack[i]->major == maParagraph && t->major == maInline) {
                    // We found a paragraph, and this is the end of parameterless inline markup.
                    // Auto-open it, pretending it has always been open.
                    openAt(i+1, t->major, t->minor);
                    stack[i+1]->children.swap(stack[i]->children);
                    close(i+1);
                } else {
                    // Cannot auto-open
                    appendText(lex.getTokenString());
                }
            }
            break;

         case maParagraph:
            // Close it if it is open.
            if (closeUntil(t->major, t->minor)) {
                open(maParagraph, miParNormal);
                lex.skipBlanks();
            } else {
                appendText(lex.getTokenString());
            }
            break;            

         case maGroup:
            if (t->minor == miGroupList) {
                // Closing a list: only close it when there actually is one open
                // FIXME: shouldn't pass Quote
                if (closeUntil(maGroup, miGroupList)) {
                    open(maParagraph, miParNormal);
                    lex.skipBlanks();
                } else {
                    // No open list.
                    appendText(lex.getTokenString());
                }
            } else if (t->minor == miGroupQuote) {
                // Closing a quote: always close
                if (closeUntil(t->major, t->minor)) {
                    open(maParagraph, miParNormal);
                    lex.skipBlanks();
                } else {
                    appendText(lex.getTokenString());
                }
            } else {
                appendText(lex.getTokenString());
            }
            break;

         default:
            appendText(lex.getTokenString());
            break;
        }
    }
}

void
BBParser::handleSmiley()
{
    open(maSpecial, miSpecialSmiley);
    stack.back()->text = lex.getTag();
    close();
}

void
BBParser::handleParagraph()
{
    closeInline();
    ASSERT(!stack.empty());
    close();
    open(maParagraph, miParNormal);
}

void
BBParser::handleAtLink()
{
    closeLinks();
    open(maLink, miLinkUser);
    stack.back()->text = lex.getAttribute();
    close();
}

void
BBParser::handleText()
{
    /* This is used for the inside of tags, where text is treated normally
       (i.e. not [code] or [noparse]).

       Although BBLexer is free to split text anywhere it seems fit,
       we assume to receive chunks of text large enough to sensibly work with.
       So far, BBLexer splits text at newlines, "[" and "@" (but "@" not after
       a letter). This is enough to detect URLs and smileys with a good chance. */
    const string_t str = lex.getTokenString();
    string_t::size_type pos = 0;
    if (!options.empty()) {
        InlineInfo info;
        while (findInlineElement(str, pos, options, info)) {
            appendText(string_t(str, pos, info.start - pos));
            switch (info.what) {
             case InlineInfo::Smiley:
                open(maSpecial, miSpecialSmiley);
                stack.back()->text = info.text;
                close();
                break;

             case InlineInfo::Link:
                // No link detection when we are inside a link!
                if (inLink()) {
                    appendText(string_t(str, info.start, info.length));
                } else {
                    open(maLink, miLinkUrl);
                    stack.back()->text = info.text;
                    close();
                }
                break;
            }
            pos = info.start + info.length;
        }
    }
    appendText(str.substr(pos));
}

void
BBParser::handleNoparse()
{
    next();                     // [noparse]
    while (current != BBLexer::Eof && (current != BBLexer::TagEnd || lex.getTag() != "noparse")) {
        appendText(lex.getTokenString());
        next();
    }
}

void
BBParser::handleCode()
{
    closeInline();
    close();                    // closes a paragraph
    open(maParagraph, miParCode);
    stack.back()->text = lex.getAttribute();
    next();                     // [code]
    while (current != BBLexer::Eof && (current != BBLexer::TagEnd || lex.getTag() != "code")) {
        appendText(lex.getTokenString());
        next();
    }
    close();
    open(maParagraph, miParNormal);
}

void
BBParser::handleListItem()
{
    depth_t i = stack.size()-1;
    while (i > 0 && (stack[i]->major != maGroup || (stack[i]->minor != miGroupListItem && stack[i]->minor != miGroupQuote))) {
        --i;
    }
    if (stack[i]->major == maGroup && stack[i]->minor == miGroupListItem) {
        // Regular case. Close list item and open new one.
        close(i);
        open(maGroup, miGroupListItem);
        open(maParagraph, miParNormal);
        lex.skipBlanks();
    } else if (stack[i]->major == maGroup && stack[i]->minor == miGroupQuote) {
        // Special case. We are inside a quote and have no list open.
        // Open one; probably the user broke the list in the middle when quoting.
        openAt(i+1, maGroup, miGroupList);
        openAt(i+2, maGroup, miGroupListItem);
        close(i+2);
        open(maGroup, miGroupListItem);
        open(maParagraph, miParNormal);
        lex.skipBlanks();
    } else {
        // No open list item
        appendText(lex.getTokenString());
    }
}

void
BBParser::next()
{
    current = lex.read();
}

void
BBParser::open(uint8_t major, uint8_t minor)
{
    TextNode* node = new TextNode();
    node->major = major;
    node->minor = minor;
    stack.push_back(node);
}

void
BBParser::openAt(ptr_vector<TextNode>::size_type n, uint8_t major, uint8_t minor)
{
    TextNode* node = new TextNode();
    node->major = major;
    node->minor = minor;
    stack.insert(stack.begin()+n, node);
}

void
BBParser::appendText(string_t what)
{
    TextNode* last = stack.back();
    if (last->major == maParagraph && last->children.empty() && what.find_first_not_of(" \t\r\n") == string_t::npos) {
        // blank string starting a paragraph, ignore
    } else if (!last->children.empty() && last->children.back()->major == maPlain) {
        last->children.back()->text += what;
    } else {
        open(maPlain, 0);
        stack.back()->text = what;
        close();
    }
}

void
BBParser::close()
{
    ASSERT(stack.size() > 1);
    TextNode* node = stack.extract_last();
    bool keep = true;
    switch (TextMajorKind(node->major)) {
     case maPlain:
        keep = !node->text.empty();
        break;
     case maInline:
        keep = !node->children.empty();
        break;
     case maInlineAttr:
        keep = !node->children.empty();
        break;
     case maLink:
        completeLink(node);
        keep = true;
        break;
     case maParagraph:
        keep = (node->minor == miParBreak || !node->children.empty());
        break;
     case maGroup:
        if (node->minor == miGroupListItem || node->minor == miGroupList) {
            keep = !node->children.empty();
        } else {
            keep = true;
        }
        break;
     case maSpecial:
        if (node->minor == miSpecialImage) {
            completeLink(node);
        }
        keep = true;
        break;
    }

    if (keep) {
        stack.back()->children.push_back(node);
    } else {
        delete node;
    }
}

void
BBParser::close(ptr_vector<TextNode>::size_type n)
{
    while (stack.size() > n) {
        close();
    }
}

/** Close all links. If the stack contains a link, closes everything
    up to and including it. */
void
BBParser::closeLinks()
{
    depth_t i = stack.size()-1;
    while (i > 0 && stack[i]->major != maParagraph && stack[i]->major != maGroup && stack[i]->major != maLink) {
        --i;
    }
    if (i > 0 && stack[i]->major == maLink) {
        close(i);
    }
}

/** Close all inline tags. Upon return, the stack contains a paragraph. */
void
BBParser::closeInline()
{
    while (!stack.empty() && stack.back()->major != maParagraph) {
        close();
    }
}

/** Close all tags up to a specific one.
    \param major,minor Tag to close
    \retval true tag found and closed
    \retval false tag not found, stack unchanged */
bool
BBParser::closeUntil(uint8_t major, uint8_t minor)
{
    depth_t i = stack.size()-1;
    while (i > 0 && (stack[i]->major != major || stack[i]->minor != minor)) {
        --i;
    }
    if (i > 0) {
        close(i);
        return true;
    } else {
        // Tag is not open
        return false;
    }
}

/** Check for link on stack. */
bool
BBParser::inLink() const
{
    depth_t i = stack.size()-1;
    while (i > 0 && stack[i]->major != maParagraph && stack[i]->major != maGroup && stack[i]->major != maLink && stack[i]->major != maSpecial) {
        --i;
    }
    return (i > 0 && (stack[i]->major == maLink || stack[i]->major == maSpecial));
}

#ifdef TEST_BBPARSE
/*
    g++-3.3 -DTEST_BBPARSE -o test_bbparse -Icpluslib -I. planetscentral/talk/bbparse.cc -L. -lcplus
 */
#include <iostream>
#include <cstdio>
#include "planetscentral/talk/bblex.cc"

static void
dump(int indent, TextNode* p)
{
    const char* what = p->major == maPlain
        ? "plain"
        : p->major == maInline
        ? (p->minor == miInBold ? "inline-bold"
           : p->minor == miInItalic ? "inline-italic"
           : p->minor == miInStrikeThrough ? "inline-strike"
           : p->minor == miInUnderline ? "inline-under"
           : p->minor == miInMonospace ? "inline-tt" : "inline-?")
        : p->major == maInlineAttr
        ? (p->minor == miIAColor ? "ia-color"
           : p->minor == miIASize ? "ia-size"
           : p->minor == miIAFont ? "ia-font" : "ia-?")
        : p->major == maLink
        ? (p->minor == miLinkUrl ? "link-url"
           : p->minor == miLinkEmail ? "link-email"
           : p->minor == miLinkThread ? "link-thread"
           : p->minor == miLinkPost ? "link-post"
           : p->minor == miLinkGame ? "link-game"
           : p->minor == miLinkUser ? "link-user"
           : p->minor == miLinkForum ? "link-forum"
           : p->minor == miLinkImage ? "link-image" : "link-?")
        : p->major == maParagraph
        ? (p->minor == miParNormal ? "paragraph"
           : p->minor == miParCode ? "code"
           : p->minor == miParCentered ? "centered" : "par-?")
        : p->major == maGroup
        ? (p->minor == miGroupRoot ? "root"
           : p->minor == miGroupQuote ? "quote"
           : p->minor == miGroupListItem ? "list-item"
           : p->minor == miGroupList ? "list" : "group-?")
        : "?";

    std::printf("%*s%s", indent, "", what);
    if (!p->text.empty()) {
        std::printf(" '%s'", p->text.c_str());
    }
    std::printf("\n");
    for (ptr_vector<TextNode>::size_type i = 0; i < p->children.size(); ++i) {
        dump(indent+4, p->children[i]);
    }
}

int main()
{
    string_t text;
    string_t line;
    while (getline(std::cin, line)) {
        if (!text.empty()) {
            text += "\n";
        }
        text += line;
    }
    BBLexer lex(text);

    TextNode* result = BBParser(lex).parse();
    dump(0, result);
    delete result;
}
#endif
