/**
  *  \file planetscentral/talk/render.h
  *  \brief Talk: Rendering
  */
#ifndef C2SERVER_PLANETSCENTRAL_TALK_RENDER_H
#define C2SERVER_PLANETSCENTRAL_TALK_RENDER_H

#include "cpluslib/string.h"
#include "int/args.h"

class DbClient;

/** Renderer state. This class stores parameters for rendering, and performs
    actual rendering.

    Rendering uses trusted and untrusted attributes. Untrusted are
    - base URL
    - format
    Those can be set with each renderer invocation.

    Trusted are:
    - message Id (message we're referring to, for "quote:" format). Caller
      makes sure that this posting is actually accessible to the user.
    - message Author (alternative to message Id to specify author for
      "quote:" format)
    - user Id (user in whose context permission checks are performed)

    Rendering transforms text in one format into another. The source text
    has the form "TYPE:TEXT", where TYPE is a supported text format. The
    output format is a TYPE, optionally prefixed by one or more transformations.
    If the both TYPEs agree, or if the output format is "raw", rendering
    is a null operation. */
class RenderState {
 public:
    RenderState();
    void clear();

    void setBaseUrl(string_t url);
    void setFormat(string_t format);
    void setMessageId(int32_t id);
    void setMessageAuthor(string_t author);
    void setUser(string_t user);

    const string_t& getBaseUrl() const;
    const string_t& getUser() const;
    int32_t getMessageId() const;
    const string_t& getMessageAuthor() const;

    bool handleOption(string_t optUC, IntArgBlock& args);
    string_t render(string_t text, DbClient& db);

 private:
    string_t baseUrl;
    string_t format;
    int32_t messageId;
    string_t messageAuthor;
    string_t user;
};

#endif
