/**
  *  \file planetscentral/talk/talk.cc
  */

#include <fcntl.h>
#include <signal.h>
#include <stdio.h>
#include <unistd.h>
#include "planetscentral/talk/talk.h"
#include "cpluslib/string.h"
#include "int/array.h"
#include "planetscentral/dbclient/dbclient.h"
#include "planetscentral/dbclient/mailclient.h"
#include "planetscentral/dbclient/reconnect.h"
#include "planetscentral/osutil.h"
#include "planetscentral/talk/talkconnection.h"
#include "server/config.h"
#include "server/net.h"
#include "util/misc.h"
#include "io/streams.h"
#include "planetscentral/talk/syntaxnames.h"
#include "planetscentral/talk/syntaxdb.h"
#include "util/except.h"

static string_t arg_listen_host = "127.0.0.1";
static string_t arg_listen_port = "5555";
static int32_t arg_threads = 5;
static string_t arg_syntax_db;
Mutex output_mutex;
DbReadyClient database_connection;
ReconnectingConnection database_rc(database_connection, output_mutex, "127.0.0.1", "6379");
MailClient mail_connection;
ReconnectingConnection mail_rc(mail_connection, output_mutex, "127.0.0.1", "21212");
string_t msgid_suffix("@localhost");
string_t path_host("localhost");
string_t www_prefix("http://localhost/");
string_t user_key("unconfigured");

static void
processConfig(string_t key, string_t value)
{
    if (key == "TALK.HOST") {
        /* @q Talk.Host:Str (Config)
           Listen address */
        arg_listen_host = value;
    } else if (key == "TALK.PORT") {
        /* @q Talk.Port:Int (Config)
           Port number. */
        arg_listen_port = value;
    } else if (key == "TALK.THREADS") {
        /* @q Talk.Threads:Int (Config)
           Number of threads (=maximum number of parallel connections) */
        stringValue(value, arg_threads);
    } else if (key == "TALK.MSGID") {
        /* @q Talk.MsgID:Str (Config)
           Suffix for creating NNTP Message-IDs.
           The value should start with a punctuator and must include a "@",
           for example, ".talk@msgid.example.com".
           The Id will be generated by prepending numbers (sequence number and {@type MID|posting Id}). */
        msgid_suffix = value;
    } else if (key == "TALK.PATH") {
        /* @q Talk.Path:Str (Config)
           Name of NNTP server, used for generating "Path" and "Xref" headers. */
        path_host = value;
    } else if (key == "TALK.WWWROOT") {
        /* @q Talk.WWWRoot:Str (Config)
           Root of web application, used for generating links. */
        www_prefix = value;
    } else if (key == "TALK.SYNTAXDB") {
        /* @q Talk.SyntaxDB:Str (Config)
           Name of file with syntax database.
           If not specified, the syntax database will be empty ({SYNTAXGET} will always fail). */
        arg_syntax_db = value;
    } else if (key == "REDIS.HOST") {
        database_rc.setHost(value);
    } else if (key == "REDIS.PORT") {
        database_rc.setPort(value);
    } else if (key == "MAILOUT.HOST") {
        mail_rc.setHost(value);
    } else if (key == "MAILOUT.PORT") {
        mail_rc.setPort(value);
    } else if (key == "USER.KEY") {
        user_key = value;
    }
}

static int server_socket;

static void*
serverLoop(void*)
{
    while (1) {
        TalkConnection conn;
        if (!conn.accept(server_socket)) {
            perror("accept");
            sleep(10);
        } else {
            try {
                while (!conn.isAtEnd()) {
                    try {
                        conn.handleRequest();
                    }
                    catch (std::exception& e) {
                        {
                            MutexGuard g(output_mutex);
                            printf("[cmd] [exception] %s\n", e.what());
                        }
                        conn.sendError(e.what());
                    }
                    conn.flush();
                }
            }
            catch (std::exception& e) {
                printf("c2talk: %s\n", e.what());
            }
            conn.close();
        }
    }
}

int main()
{
    /* Unbuffer stdout */
    setvbuf(stdout, 0, _IOLBF, 0);

    /* Configuration */
    loadConfiguration(processConfig);
    if (!arg_syntax_db.empty()) {
        try { 
            Ptr<Stream> s = Stream::openFile(arg_syntax_db, Stream::C_READ);
            syntax_db.load(*s);
        }
        catch (FileProblemException& e) {
            fprintf(stderr, "unable to load syntax db: %s: %s\n", e.getFileName().c_str(), e.what());
            return 1;
        }
    }

    /* Connect to database */
    try {
        database_rc.connectRetry();
        mail_rc.connectRetry();
    }
    catch (std::exception& e) {
        fprintf(stderr, "unable to connect to servers: %s\n", e.what());
        return 1;
    }

    /* Listen */
    server_socket = Bind().withAddress(arg_listen_host, arg_listen_port);
    if (server_socket < 0) {
        fprintf(stderr, "[main] unable to bind socket\n");
        return 1;
    }

    /* OK, main loop */
    printf("[main] listening...\n");
    signal(SIGPIPE, SIG_IGN);
    fcntl(server_socket, F_SETFD, FD_CLOEXEC);
    listen(server_socket, 10);

    /* Create some threads. Create one less than configured; the main
       thread will also serve requests. */
    for (int32_t i = 1; i < arg_threads; ++i) {
        pthread_t handle;
        if (pthread_create(&handle, 0, serverLoop, 0) < 0) {
            perror("[main] unable to create thread");
            return 1;
        }
    }
    serverLoop(0);
    close(server_socket);
}
