/**
  *  \file planetscentral/talk/textout.cc
  */

#include "planetscentral/talk/textout.h"
#include "cpluslib/string.h"
#include "planetscentral/dbclient/dbsubtree.h"
#include "util/misc.h"
#include "cpluslib/utf8.h"
#include "planetscentral/talk/text.h"
#include "planetscentral/talk/db.h"
#include "planetscentral/talk/forum.h"
#include "planetscentral/talk/message.h"
#include "planetscentral/talk/topic.h"
#include "planetscentral/talk/render.h"

static string_t
hackSubject(const string_t s)
{
    if (s.empty()) {
        return "(no subject)";
    } else if (s.size() > 30) {
        std::size_t i = 28;
        while (i > 0 && (s[i-1] == ' ' || Utf8Reader::isContinuationByte(s[i]))) {
            // Yes, this is intended to check i-1 for spaces, and i for continuation
            --i;
        }
        return string_t(s, 0, i) + "...";
    } else {
        return s;
    }
}

static string_t
makeGameName(const string_t& text, RenderState& state, DbClient& db)
{
    // Parse game Id
    int32_t gameId;
    if (!stringValue(text, gameId) || gameId <= 0) {
        return text;
    }

    // Access game, check permissions
    // xref host/game.cc, Game::hasPermission
    DbSubtree root(db, GAME_ROOT);
    if (!root.intSetKey("all").contains(gameId)) {
        return text;
    }

    DbSubtree game(root.subtree(gameId));
    const string_t gameState = game.stringKey("state").get();
    if (gameState != "joining" && gameState != "running" && gameState != "finished") {
        return text;
    }

    const string_t gameType = game.stringKey("type").get();
    if (gameType != "unlisted"
        && gameType != "public"
        && game.stringKey("owner").get() != state.getUser()
        && !game.hashKey("users").hasField(state.getUser()))
    {
        return text;
    }

    // OK, we are allowed to access it. Get its name.
    return game.stringKey("name").get();
}

static string_t
makeForumName(const string_t& text, RenderState& /*state*/, DbClient& db)
{
    // Parse forum Id
    int32_t forumId;
    if (!stringValue(text, forumId) || forumId <= 0) {
        return text;
    }

    // Access forum, check permissions
    Forum forum(db, forumId);
    if (!forum.exists()) {
        return text;
    }

    // OK, we are allowed to access it. Get its name.
    return forum.name().get();
}

static string_t
makePostName(const string_t& text, RenderState& /*state*/, DbClient& db)
{
    // Validate
    int32_t messageId;
    if (!stringValue(text, messageId) || messageId <= 0) {
        return text;
    }

    // Access thread, check permissions
    Message m(db, messageId);
    if (!m.exists()) {
        // FIXME permission check!
        return text;
    }
    int32_t topicId = m.topicId().get();
    Topic t(db, topicId);

    // OK, we are allowed to access it. Get its name.
    return hackSubject(m.subject().get());
}

static string_t
makeThreadName(const string_t& text, RenderState& /*state*/, DbClient& db)
{
    // Validate
    int32_t topicId;
    if (!stringValue(text, topicId) || topicId <= 0) {
        return text;
    }

    // Access thread, check permissions
    Topic t(db, topicId);
    if (!t.exists()) {
        // FIXME permission check!
        return text;
    }

    // OK, we are allowed to access it. Get its name.
    return hackSubject(t.subject().get());
}

/** Render node as plaintext.
    This is a version of TextNode::rawTextContent(),
    but it fills in game names etc. */
string_t
renderText(TextNode* node, RenderState& state, DbClient& db)
{
    if (node->major == maPlain) {
        return node->text;
    } else {
        string_t result;
        for (ptr_vector<TextNode>::size_type i = 0, n = node->children.size(); i < n && result.size() < 10000; ++i) {
            result += renderText(node->children[i], state, db);
        }
        if (result.empty() && node->major == maLink) {
            switch (TextLinkFormat(node->minor)) {
             case miLinkUrl:
             case miLinkEmail:
             case miLinkUser:
                // Fallback is using the text
                result = node->text;
                break;

             case miLinkThread:
                result = makeThreadName(node->text, state, db);
                break;

             case miLinkPost:
                result = makePostName(node->text, state, db);
                break;

             case miLinkGame:
                result = makeGameName(node->text, state, db);
                break;

             case miLinkForum:
                result = makeForumName(node->text, state, db);
                break;
            }
        }
        return result;
    }
}
